@@ -1,8 +1,8 @@
 /* Pango
- * pangatsui-fontmap.c
+ * pangoatsui-fontmap.c
  *
  * Copyright (C) 2000-2003 Red Hat, Inc.
- * Copyright (C) 2005 Imendio AB
+ * Copyright (C) 2005-2007 Imendio AB
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -24,13 +24,11 @@
 
 #include "pango-fontmap.h"
 #include "pangoatsui-private.h"
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 #include "modules.h"
 
 #import <Cocoa/Cocoa.h>
 
-typedef struct _PangoATSUIFamily PangoATSUIFamily;
-typedef struct _PangoATSUIFace   PangoATSUIFace;
 typedef struct _FontHashKey      FontHashKey;
 
 struct _PangoATSUIFamily
@@ -45,11 +43,11 @@ struct _PangoATSUIFamily
 
 #define PANGO_TYPE_ATSUI_FAMILY              (pango_atsui_family_get_type ())
 #define PANGO_ATSUI_FAMILY(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_ATSUI_FAMILY, PangoATSUIFamily))
-#define PANGO_ATSUI_IS_FAMILY(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_ATSUI_FAMILY))
+#define PANGO_IS_ATSUI_FAMILY(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_ATSUI_FAMILY))
 
 #define PANGO_TYPE_ATSUI_FACE              (pango_atsui_face_get_type ())
 #define PANGO_ATSUI_FACE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_ATSUI_FACE, PangoATSUIFace))
-#define PANGO_ATSUI_IS_FACE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_ATSUI_FACE))
+#define PANGO_IS_ATSUI_FACE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_ATSUI_FACE))
 
 struct _PangoATSUIFace
 {
@@ -57,11 +55,14 @@ struct _PangoATSUIFace
 
   PangoATSUIFamily *family;
 
+  PangoCoverage *coverage;
+
   char *postscript_name;
   char *style_name;
 
   int weight;
   int traits;
+  guint synthetic_italic : 1;
 };
 
 static GType pango_atsui_family_get_type (void);
@@ -98,18 +99,30 @@ pango_atsui_family_list_faces (PangoFontFamily  *family,
 			       int              *n_faces)
 {
   PangoATSUIFamily *atsuifamily = PANGO_ATSUI_FAMILY (family);
-  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
 
-  if (atsuifamily->n_faces < 0) 
+  if (atsuifamily->n_faces < 0)
     {
+      NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
       const char *real_family = get_real_family (atsuifamily->family_name);
-      NSArray *members = [[NSFontManager sharedFontManager] availableMembersOfFontFamily:[NSString stringWithUTF8String:real_family]];
-      int i;
+      NSFontManager *manager = [NSFontManager sharedFontManager];
+      NSArray *members = [manager availableMembersOfFontFamily:[NSString stringWithUTF8String:real_family]];
+      int i, count;
+      GHashTable *hash_table;
+      GList *faces = NULL, *l;
+      GList *synthetic_faces = NULL;
 
-      atsuifamily->n_faces = [members count];
-      atsuifamily->faces = g_new (PangoFontFace *, atsuifamily->n_faces);
+      /* The NSFontManager API returns italic faces for some families
+       * even if they don't exist. When using Cocoa to create
+       * instances of those fonts, Cocoa synthesizes italic versions
+       * by applying a shear transformation. We do that manually for
+       * those fonts in pangocairo-atsuifont.c. For many other fonts,
+       * there is no italic face at all, so we create synthesized
+       * versions of those like in the win32 and fontconfig backends.
+       */
+      hash_table = g_hash_table_new (g_direct_hash, g_direct_equal);
 
-      for (i = 0; i < atsuifamily->n_faces; i++)
+      count = [members count];
+      for (i = 0; i < count; i++)
 	{
 	  PangoATSUIFace *face = g_object_new (PANGO_TYPE_ATSUI_FACE, NULL);
 	  NSArray *font_array = [members objectAtIndex:i];
@@ -120,8 +133,48 @@ pango_atsui_family_list_faces (PangoFontFamily  *family,
 	  face->weight = [[font_array objectAtIndex:2] intValue];
 	  face->traits = [[font_array objectAtIndex:3] intValue];
 
-	  atsuifamily->faces[i] = (PangoFontFace *)face;
+	  faces = g_list_prepend (faces, face);
+
+	  if (face->traits & NSItalicFontMask)
+	    g_hash_table_insert (hash_table, GINT_TO_POINTER (face->weight), face);
 	}
+
+      for (l = faces; l; l = l->next)
+	{
+	  PangoATSUIFace *face = l->data;
+
+	  if (!g_hash_table_lookup (hash_table, GINT_TO_POINTER (face->weight)))
+	    {
+	      PangoATSUIFace *italic_face = g_object_new (PANGO_TYPE_ATSUI_FACE, NULL);
+
+	      italic_face->family = atsuifamily;
+	      italic_face->postscript_name = g_strdup (face->postscript_name);
+	      italic_face->weight = face->weight;
+	      italic_face->traits = face->traits | NSItalicFontMask;
+	      italic_face->synthetic_italic = TRUE;
+
+	      /* Try to create a sensible face name. */
+	      if (strcasecmp (face->style_name, "regular") == 0)
+		italic_face->style_name = g_strdup ("Oblique");
+	      else 
+		italic_face->style_name = g_strdup_printf ("%s Oblique", face->style_name);
+
+	      synthetic_faces = g_list_prepend (synthetic_faces, italic_face);
+	    }
+	}
+
+      faces = g_list_concat (faces, synthetic_faces);
+
+      atsuifamily->n_faces = g_list_length (faces);
+      atsuifamily->faces = g_new (PangoFontFace *, atsuifamily->n_faces);
+
+      for (l = faces, i = 0; l; l = l->next, i++)
+	atsuifamily->faces[i] = l->data;
+
+      g_list_free (faces);
+      g_hash_table_destroy (hash_table);
+
+      [pool release];
     }
 
   if (n_faces)
@@ -129,8 +182,6 @@ pango_atsui_family_list_faces (PangoFontFamily  *family,
 
   if (faces)
     *faces = g_memdup (atsuifamily->faces, atsuifamily->n_faces * sizeof (PangoFontFace *));
-
-  [pool release];
 }
 
 static const char *
@@ -156,13 +207,11 @@ pango_atsui_family_finalize (GObject *object)
 
   g_free (family->family_name);
 
-  if (family->n_faces != -1) 
+  if (family->n_faces != -1)
     {
-      for (i = 0; i < family->n_faces; i++) 
-	{
-	  g_object_unref (family->faces[i]);
-	}
-      
+      for (i = 0; i < family->n_faces; i++)
+	g_object_unref (family->faces[i]);
+
       g_free (family->faces);
     }
 
@@ -174,7 +223,7 @@ pango_atsui_family_class_init (PangoFontFamilyClass *class)
 {
   GObjectClass *object_class = (GObjectClass *)class;
   int i;
-  
+
   pango_atsui_family_parent_class = g_type_class_peek_parent (class);
 
   object_class->finalize = pango_atsui_family_finalize;
@@ -198,26 +247,26 @@ pango_atsui_family_get_type (void)
 {
   static GType object_type = 0;
 
-  if (!object_type)
+  if (G_UNLIKELY (!object_type))
     {
-      static const GTypeInfo object_info =
+      const GTypeInfo object_info =
       {
-        sizeof (PangoFontFamilyClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_atsui_family_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoATSUIFamily),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) pango_atsui_family_init,
+	sizeof (PangoFontFamilyClass),
+	(GBaseInitFunc) NULL,
+	(GBaseFinalizeFunc) NULL,
+	(GClassInitFunc) pango_atsui_family_class_init,
+	NULL,           /* class_finalize */
+	NULL,           /* class_data */
+	sizeof (PangoATSUIFamily),
+	0,              /* n_preallocs */
+	(GInstanceInitFunc) pango_atsui_family_init,
       };
-      
+
       object_type = g_type_register_static (PANGO_TYPE_FONT_FAMILY,
-                                            I_("PangoATSUIFamily"),
-                                            &object_info, 0);
+					    I_("PangoATSUIFamily"),
+					    &object_info, 0);
     }
-  
+
   return object_type;
 }
 
@@ -230,7 +279,7 @@ pango_atsui_face_describe (PangoFontFace *face)
   PangoStyle pango_style;
   PangoVariant pango_variant;
   int weight;
-  
+
   description = pango_font_description_new ();
 
   pango_font_description_set_family (description, atsuiface->family->family_name);
@@ -251,7 +300,7 @@ pango_atsui_face_describe (PangoFontFace *face)
     pango_weight = PANGO_WEIGHT_ULTRABOLD;
   else if (weight == 13 || weight == 14)
     pango_weight = PANGO_WEIGHT_HEAVY;
-  else 
+  else
     g_assert_not_reached ();
 
   if (atsuiface->traits & NSItalicFontMask)
@@ -266,8 +315,9 @@ pango_atsui_face_describe (PangoFontFace *face)
 
   pango_font_description_set_weight (description, pango_weight);
   pango_font_description_set_style (description, pango_style);
+  pango_font_description_set_variant (description, pango_variant);
 
-  return description;  
+  return description;
 }
 
 static const char *
@@ -280,8 +330,8 @@ pango_atsui_face_get_face_name (PangoFontFace *face)
 
 static void
 pango_atsui_face_list_sizes (PangoFontFace  *face,
-                             int           **sizes,
-                             int            *n_sizes)
+			     int           **sizes,
+			     int            *n_sizes)
 {
   *n_sizes = 0;
   *sizes = NULL;
@@ -292,12 +342,23 @@ pango_atsui_face_finalize (GObject *object)
 {
   PangoATSUIFace *atsuiface = PANGO_ATSUI_FACE (object);
 
+  if (atsuiface->coverage)
+    pango_coverage_unref (atsuiface->coverage);
+
   g_free (atsuiface->postscript_name);
   g_free (atsuiface->style_name);
 
   G_OBJECT_CLASS (pango_atsui_face_parent_class)->finalize (object);
 }
 
+static gboolean
+pango_atsui_face_is_synthesized (PangoFontFace *face)
+{
+  PangoATSUIFace *atsuiface = PANGO_ATSUI_FACE (face);
+
+  return atsuiface->synthetic_italic;
+}
+
 static void
 pango_atsui_face_class_init (PangoFontFaceClass *class)
 {
@@ -310,6 +371,7 @@ pango_atsui_face_class_init (PangoFontFaceClass *class)
   class->describe = pango_atsui_face_describe;
   class->get_face_name = pango_atsui_face_get_face_name;
   class->list_sizes = pango_atsui_face_list_sizes;
+  class->is_synthesized = pango_atsui_face_is_synthesized;
 }
 
 GType
@@ -317,29 +379,63 @@ pango_atsui_face_get_type (void)
 {
   static GType object_type = 0;
 
-  if (!object_type)
+  if (G_UNLIKELY (!object_type))
     {
-      static const GTypeInfo object_info =
+      const GTypeInfo object_info =
       {
-        sizeof (PangoFontFaceClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_atsui_face_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoATSUIFace),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) NULL,
+	sizeof (PangoFontFaceClass),
+	(GBaseInitFunc) NULL,
+	(GBaseFinalizeFunc) NULL,
+	(GClassInitFunc) pango_atsui_face_class_init,
+	NULL,           /* class_finalize */
+	NULL,           /* class_data */
+	sizeof (PangoATSUIFace),
+	0,              /* n_preallocs */
+	(GInstanceInitFunc) NULL,
       };
-      
+
       object_type = g_type_register_static (PANGO_TYPE_FONT_FACE,
-                                            I_("PangoATSUIFace"),
-                                            &object_info, 0);
+					    I_("PangoATSUIFace"),
+					    &object_info, 0);
     }
-  
+
   return object_type;
 }
 
+const char *
+_pango_atsui_face_get_postscript_name (PangoATSUIFace *face)
+{
+  return face->postscript_name;
+}
+
+gboolean
+_pango_atsui_face_get_synthetic_italic (PangoATSUIFace *face)
+{
+  return face->synthetic_italic;
+}
+
+PangoCoverage *
+_pango_atsui_face_get_coverage (PangoATSUIFace *face,
+				PangoLanguage  *language)
+{
+  int i;
+
+  /* Note: We currently fake the coverage by reporting that the 255 first 
+   * glyphs exist in all fonts, which is not true. It's unclear how to get
+   * this done correctly.
+   */
+
+  if (face->coverage)
+    return face->coverage;
+
+  face->coverage = pango_coverage_new ();
+
+  for (i = 0; i < 256; i++)
+    pango_coverage_set (face->coverage, i, PANGO_COVERAGE_EXACT);
+
+  return face->coverage;
+}
+
 static void pango_atsui_font_map_class_init (PangoATSUIFontMapClass *class);
 static void pango_atsui_font_map_init (PangoATSUIFontMap *atsuifontmap);
 
@@ -354,7 +450,7 @@ static void
 pango_atsui_font_map_finalize (GObject *object)
 {
   PangoATSUIFontMap *fontmap = PANGO_ATSUI_FONT_MAP (object);
-  
+
   g_hash_table_destroy (fontmap->font_hash);
   g_hash_table_destroy (fontmap->families);
 
@@ -370,7 +466,7 @@ struct _FontHashKey {
 };
 
 /* Fowler / Noll / Vo (FNV) Hash (http://www.isthe.com/chongo/tech/comp/fnv/)
- * 
+ *
  * Not necessarily better than a lot of other hashes, but should be OK, and
  * well tested with binary data.
  */
@@ -388,7 +484,7 @@ hash_bytes_fnv (unsigned char *buffer,
       hval *= FNV_32_PRIME;
       hval ^= *buffer++;
     }
-  
+
   return hval;
 }
 
@@ -439,7 +535,7 @@ font_hash_key_free (FontHashKey *key)
   if (key->context_key)
     PANGO_ATSUI_FONT_MAP_GET_CLASS (key->fontmap)->context_key_free (key->fontmap,
 								     key->context_key);
-  
+
   g_slice_free (FontHashKey, key);
 }
 
@@ -447,7 +543,7 @@ static FontHashKey *
 font_hash_key_copy (FontHashKey *old)
 {
   FontHashKey *key = g_slice_new (FontHashKey);
-  
+
   key->fontmap = old->fontmap;
   key->matrix = old->matrix;
   key->desc = pango_font_description_copy (old->desc);
@@ -457,7 +553,7 @@ font_hash_key_copy (FontHashKey *old)
 											old->context_key);
   else
     key->context_key = NULL;
-      
+
   return key;
 }
 
@@ -487,7 +583,7 @@ font_hash_key_for_context (PangoATSUIFontMap *fcfontmap,
 {
   key->fontmap = fcfontmap;
   get_context_matrix (context, &key->matrix);
-  
+
   if (PANGO_ATSUI_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get)
     key->context_key = (gpointer)PANGO_ATSUI_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get (fcfontmap, context);
   else
@@ -501,36 +597,67 @@ pango_atsui_font_map_add (PangoATSUIFontMap *atsuifontmap,
 {
   FontHashKey key;
   FontHashKey *key_copy;
+  PangoATSUIFace *face;
 
-  g_assert (atsuifont->fontmap == NULL);
-
-  atsuifont->fontmap = g_object_ref (atsuifontmap);
+  _pango_atsui_font_set_font_map (atsuifont, atsuifontmap);
 
   font_hash_key_for_context (atsuifontmap, context, &key);
-  key.postscript_name = atsuifont->postscript_name;
-  key.desc = atsuifont->desc;
+  face = _pango_atsui_font_get_face (atsuifont);
+  key.postscript_name = (char *)_pango_atsui_face_get_postscript_name (face);
+  key.desc = _pango_atsui_font_get_font_description (atsuifont);
 
   key_copy = font_hash_key_copy (&key);
-  atsuifont->context_key = key_copy->context_key;
-  atsuifont->matrix = key.matrix;
+  _pango_atsui_font_set_context_key (atsuifont, key_copy->context_key);
   g_hash_table_insert (atsuifontmap->font_hash, key_copy, g_object_ref (atsuifont));
 }
 
 static PangoATSUIFont *
-pango_atsui_font_map_lookup (PangoATSUIFontMap *atsuifontmap,
-			     PangoContext      *context,
+pango_atsui_font_map_lookup (PangoATSUIFontMap    *atsuifontmap,
+			     PangoContext         *context,
 			     PangoFontDescription *desc,
-			     const char        *postscript_name)
+			     PangoATSUIFace       *face)
 {
   FontHashKey key;
-  
+
   font_hash_key_for_context (atsuifontmap, context, &key);
-  key.postscript_name = (char *)postscript_name;
+  key.postscript_name = (char *)_pango_atsui_face_get_postscript_name (face);
   key.desc = desc;
 
   return g_hash_table_lookup (atsuifontmap->font_hash, &key);
 }
 
+static gboolean
+find_best_match (PangoATSUIFamily            *font_family,
+		 const PangoFontDescription  *description,
+		 PangoFontDescription       **best_description,
+		 PangoATSUIFace             **best_face)
+{
+  PangoFontDescription *new_desc;
+  int i;
+
+  *best_description = NULL;
+  *best_face = NULL;
+
+  for (i = 0; i < font_family->n_faces; i++)
+    {
+      new_desc = pango_font_face_describe (font_family->faces[i]);
+
+      if (pango_font_description_better_match (description, *best_description, new_desc))
+	{
+	  pango_font_description_free (*best_description);
+	  *best_description = new_desc;
+	  *best_face = (PangoATSUIFace *)font_family->faces[i];
+	}
+      else
+	pango_font_description_free (new_desc);
+    }
+
+  if (*best_description)
+    return TRUE;
+
+  return FALSE;
+}
+
 static PangoFont *
 pango_atsui_font_map_load_font (PangoFontMap               *fontmap,
 				PangoContext               *context,
@@ -541,62 +668,53 @@ pango_atsui_font_map_load_font (PangoFontMap               *fontmap,
   gchar *name;
   gint size;
 
-  g_return_val_if_fail (description != NULL, NULL);
-
   size = pango_font_description_get_size (description);
-
-  if (size <= 0)
+  if (size < 0)
     return NULL;
 
   name = g_utf8_casefold (pango_font_description_get_family (description), -1);
-
   font_family = g_hash_table_lookup (atsuifontmap->families, name);
   g_free (name);
+
   if (font_family)
     {
-      PangoFontDescription *best_desc = NULL, *new_desc;
-      PangoATSUIFace *best_face = NULL;
+      PangoFontDescription *best_description;
+      PangoATSUIFace *best_face;
       PangoATSUIFont *best_font;
-      int i;
-      
+
       /* Force a listing of the available faces */
       pango_font_family_list_faces ((PangoFontFamily *)font_family, NULL, NULL);
 
-      for (i = 0; i < font_family->n_faces; i++)
-	{
-	  new_desc = pango_font_face_describe (font_family->faces[i]);
-	  
-	  if (pango_font_description_better_match (description, best_desc, new_desc))
-	    {
-	      pango_font_description_free (best_desc);
-	      best_desc = new_desc;
-	      best_face = (PangoATSUIFace *)font_family->faces[i];
-	    }
-	  else 
-	    {
-	      pango_font_description_free (new_desc);
-	    }
-	}
+      if (!find_best_match (font_family, description, &best_description, &best_face))
+	return NULL;
       
-      pango_font_description_set_size (best_desc, size);
+      pango_font_description_set_size (best_description, size);
+
+      best_font = pango_atsui_font_map_lookup (atsuifontmap, 
+					       context, 
+					       best_description, 
+					       best_face);
 
-      best_font = pango_atsui_font_map_lookup (atsuifontmap, context, best_desc, best_face->postscript_name);
-      
       if (best_font)
 	g_object_ref (best_font);
       else
 	{
-	  best_font = (* PANGO_ATSUI_FONT_MAP_GET_CLASS (atsuifontmap)->create_font) (atsuifontmap, context,
-										      best_face->postscript_name, best_desc);
+	  PangoATSUIFontMapClass *klass;
 
+	  klass = PANGO_ATSUI_FONT_MAP_GET_CLASS (atsuifontmap);
+	  best_font = klass->create_font (atsuifontmap, context, 
+					  best_face, best_description);
+	  
 	  if (best_font)
 	    pango_atsui_font_map_add (atsuifontmap, context, best_font);
 	    /* TODO: Handle the else case here. */
 	}
 
-      pango_font_description_free (best_desc);
-      return PANGO_FONT (best_font);
+      pango_font_description_free (best_description);
+
+      return (PangoFont *)best_font;
     }
+
   return NULL;
 }
 
@@ -629,7 +747,7 @@ pango_atsui_font_map_list_families (PangoFontMap      *fontmap,
   if (families)
     {
       int i = 0;
-	
+
       *families = g_new (PangoFontFamily *, *n_families);
 
       tmp_list = family_list;
@@ -640,11 +758,11 @@ pango_atsui_font_map_list_families (PangoFontMap      *fontmap,
 	  tmp_list = tmp_list->next;
 	}
     }
-  
+
   g_slist_free (family_list);
 }
 
-static void 
+static void
 pango_atsui_font_map_init (PangoATSUIFontMap *atsuifontmap)
 {
   NSArray *family_array;
@@ -656,7 +774,7 @@ pango_atsui_font_map_init (PangoATSUIFontMap *atsuifontmap)
   atsuifontmap->families = g_hash_table_new_full (g_str_hash, g_str_equal,
 						  g_free, g_object_unref);
 
-  
+
   atsuifontmap->font_hash = g_hash_table_new_full ((GHashFunc)font_hash_key_hash,
 						   (GEqualFunc)font_hash_key_equal,
 						   (GDestroyNotify)font_hash_key_free,
